# ==========================================================
# Copyright (C) 2005 Z80-OOLONG. (NAKATSUKA, Yukitaka)
# ==========================================================

require 'vikiwikisys'
require 'xmlrpc/server'

module VikiWiki
	class XMLRPC < ::XMLRPC::CGIServer
		INTERFACE = ::XMLRPC::interface("wiki") do
			meth "array getRecentChanges(dateTime.iso8601)"
			meth "int getRPCVersionSupported()"
			meth "string getPage(string)"
			meth "string getPageVersion(string, int)"
			meth "string getPageHTML(string)"
			meth "string getPageHTMLVersion(string, int)"
			meth "string getPageXML(string)"
			meth "string getPageXMLVersion(string, int)"
			meth "array getAllPages()"
			meth "struct getPageInfo(string)"
			meth "struct getPageInfoVersion(string, int)"
			meth "array listLinks(string)"
			meth "array getBackLinks(string)"
			meth "boolean putPage(string, string, struct)"
		end

		def initialize(filename = nil, conf = nil)
			@sys = nil
			if (filename && conf) then
				super(".")
				@filename = filename; @conf = conf
				add_handler(INTERFACE, self)
				add_introspection
			end
		end

		def serve
			if (@filename && @conf) then
				(Thread.start do; ($SAFE < 1) && ($SAFE += 1)
					cgi = ::CGI.new("empty")
					sys_setup(@filename, @conf, cgi)
					super
				end).join
			end
		end

		def sys_setup(filename, conf, cgi)
			cgi.params.clear
			@sys = VikiWiki::Sys.new(filename, conf, cgi)
			@sys.setup
		end			

		def getRecentChanges(timestamp)
			time = timestamp.to_time; result = Array::new
			@sys.pages.each do |page|
				his = @sys.cvs.getHistory(page.path)
				his.find_all {|v, f, t| t >= time}.each do |ver|
					result << ver2struct(page.name, ver)
				end
			end
			return result.sort do |a,b|
				a["lastModified"].to_time <=> b["lastModified"].to_time
			end
		end

		def getRPCVersionSupported()
			return 2
		end

		def getPage(pagename)
			return getPageVersion(pagename, nil)
		end

		def getPageVersion(pagename, version)
			check_exist_page(pagename = u2code(pagename))
			return code2u(@sys.pages[pagename].load(version))
		end

		def getPageHTML(pagename)
			return getPageHTMLVersion(pagename, nil)
		end

		def getPageHTMLVersion(pagename, version)
			return get_http_body(pagename, version, false)
		end

		def getPageXML(pagename)
			return getPageXMLVersion(pagename, nil)
		end

		def getPageXMLVersion(pagename, version)
			return get_http_body(pagename, version, true)
		end

		def getAllPages()
			return @sys.pages.names.collect do |pagename|
				code2u(pagename)
			end
		end

		def getPageInfo(pagename)
			return getPageInfoVersion(pagename, nil)
		end

		def getPageInfoVersion(pagename, version)
			check_exist_page(pagename = u2code(pagename))
			page = @sys.pages[pagename]
			his = @sys.cvs.getHistory(page.path)
			ver = his.find {|v, f, t| v == version} || his.last
			return ver2struct(page.name, ver)
		end

		def listLinks(pagename)
			check_exist_page(pagename = u2code(pagename))
			page = @sys.pages[pagename]
			result = Array::new
			page.node.each_node(/WikiName|a/) do |node|
				struct = Hash::new
				case node.nodeName
				when "WikiName"
					name = node.getAttribute('wri')
					if name then
						struct["page"] = code2u(name)
						struct["href"] = WRI::new(@sys, name).uri
						struct["type"] = "local"
						result << struct
					end
				when "a"
					uri = node.getAttribute('href')
					if uri then
						struct["href"] = @sys.relativeuri(uri)
						unless /^(https?|ftp|mailto):/ === uri then
							uri = @sys.fulluri(uri)
						end
						struct["page"] = uri
						struct["type"] = "external"
						result << struct
					end
				end
			end
			return result
		end

		def getBackLinks(pagename)
			links = @sys.links[u2code(pagename)] || []
			return links.collect {|refer| code2u(refer)}
		end

		def putPage(pagename, content, attributes)
			page = @sys.pages[u2code(pagename)]
			path = page.path
			mtime = File::exist?(path) ? File::mtime(path) : nil
			static = attributes["statichtml"] ? true : false
			als = u2code(attributes["alias"])
			page.write(u2code(content), mtime, static)
			if als then
				@sys.aliases[page.name] = als
				@sys.aliases.update
			end
			return true
		end

		private

		def u2code(str)
			return USEconv::cnvmethod("u", @sys.code).call(str)
		end

		def code2u(str)
			return USEconv::cnvmethod(@sys.code, "u").call(str)
		end

		def check_exist_page(pagename)
			unless getAllPages().include?(pagename) then
				raise ::XMLRPC::FaultException.new(1, "No such page was found.")
			end
		end

		def get_http_body(pagename, version, xml = false)
			@sys.cgi.params["p"] = [pagename]
			xml && (@sys.cgi.params["m"] = ["XML"])
			version && (@sys.cgi.params["v"] = [version.to_s])
			@sys.setup
			body = @sys.getHTTPbody
			body.gsub!(/encoding="#{@sys.encoding}"/, 'encoding="utf-8"')
			return code2u(body)
			
		end

		def time2xmlrpcdate(time)
			return ::XMLRPC::DateTime.new(
				time.year, time.month, time.day,
				time.hour, time.min, time.sec
			)
		end

		def ver2struct(page, ver)
			result = {
				"name" 		=> code2u(page),
				"lastModified"	=> time2xmlrpcdate(ver[2]),
				"version"	=> ver[0]
			}
			ver[1] && (result["author"] = code2u(ver[1]))
			return result
		end
	end
end

# ==========================================================
