# ==========================================================
# VikiWiki::WEBrick::Servlet -- WEBrick interface of VikiWiki.
# Copyright (C) 2005 Z80-OOLONG. (NAKATSUKA, Yukitaka)
# ==========================================================

require "fileutils"
require "find"
require "webrick"
require "webrick/cgi"
require "thread"
require "xmlrpc/server"

require "vikiwiki/config"

$:.unshift($VIKIWIKIDIR)
$:.unshift("#{$VIKIWIKIDIR}/lib")

require "extension/webrick"
require "extension/stringio"
require "extension/xmlrpc"

require "vikiwikiweb"
require "vikiwikipluginhook"
require "vikiwikitrackback"
require "vikiwikixmlrpc"

module VikiWiki
	module WEBrick
		############
		# Utility module

		module Util
			############
			# sanitize and expand path

			def expand_path(path)
				if (/[|&;`$<>]/ =~ path) then
					raise ::WEBrick::HTTPStatus::InternalServerError
				end
				path = File::expand_path(path.untaint); path.gsub!(%r|^/+|, "/")
				return path
			end

			def normalize_uri(uri)
				uri = uri.gsub(%r|/+|, "/"); uri.gsub!(%r|/$|, "")
				return uri
			end

			module_function :expand_path, :normalize_uri

			############
			# definition of constant

			BASEDIR		= expand_path("#{ENV['HOME']}/VikiWiki")
			BASEURI		= "/"
			CONFIGFILE	= "#{BASEDIR}/vikiwiki.conf"
			SCRIPT_NAME	= "/vikiwiki"
			STATICDIR	= "#{BASEDIR}/static"
			STATICURI	= "/static"
			XMLRPC_CGI	= "/xmlrpc"
		end

		class Arguments
			ARGNAMES = [
				"BASEDIR", "CONFIGFILE",
				"SCRIPT_NAME", "PLUGIN_HOOK_SCRIPT",
				"TB_CGI", "XMLRPC_CGI", "PORT"
			]

			def initialize(args = {})
				@args = args.inject({}) do |a, (n, v)|
					n = n.to_s.upcase
					if ARGNAMES.member?(n) && v then
						a[n] = v
					end; a
				end
			end

			def [](name)
				check_argname(name = name.to_s.upcase)
				return @args[name]
			end
			def []=(name, value)
				check_argname(name = name.to_s.upcase)
				value && (@args[name] = value)
			end
			def delete(name)
				check_argname(name = name.to_s.upcase)
				@args.delete(name)
			end

			def to_hash; @args.dup; end
			def to_args; self; end
			class ::Hash
				def to_args
					VikiWiki::WEBrick::Arguments.new(self)
				end
			end

			private

			def check_argname(name)
				if !ARGNAMES.member?(name) then
					raise "Invalid Argument Key."
				end
			end
		end

		############
		# class Config. This class is for VikiWiki configuration

		class Config
			include Util

			############
			# Default configuration.
			# See also vikiwiki/parameter.txt in vikiwiki.zip

			DEFAULT_CONFIG = VikiWiki::CONF.clone
			DEFAULT_CONFIG["AUTHOR"]	= "Administrator"
			DEFAULT_CONFIG["ADMIN"]		= "admin"
			DEFAULT_CONFIG["ADDRESS"]	= "admin@example.com"
			DEFAULT_CONFIG["AUTHTYPE"]	= "vikiwiki"
			DEFAULT_CONFIG["BASEDIR"]	= BASEDIR
			DEFAULT_CONFIG["BASEURI"]	= BASEURI
			DEFAULT_CONFIG["LOGFILTER"]	= "all"
			DEFAULT_CONFIG["ENCODING"]	= $VIKIWIKIENCODE
			DEFAULT_CONFIG["SCRIPT_NAME"]	= SCRIPT_NAME
			DEFAULT_CONFIG["STATICDIR"]	= STATICDIR
			DEFAULT_CONFIG["STATICURI"]	= STATICURI
			DEFAULT_CONFIG["WIKIFARM"]	= false
			DEFAULT_CONFIG["TITLE"]		= "VikiWiki - page"
			DEFAULT_CONFIG["STATICTITLE"]	= "alias"
			DEFAULT_CONFIG["CONFIGFILE"]	= CONFIGFILE
			DEFAULT_CONFIG["XMLRPC_CGI"]	= XMLRPC_CGI

			END_SDEF = Regexp::escape("if $0 == __FILE__ then\n")
			BGN_UDEF = Regexp::escape(VikiWiki::Sys::BGN_UDEF)
			END_UDEF = Regexp::escape(VikiWiki::Sys::END_UDEF)

			def initialize(args)
				args = args.to_hash
				@scriptnames = args.reject do |n, v|
					/^(BASEDIR|CONFIGFILE)$/ === n
				end
				conf = init_conf(args)
				@basedir = conf["BASEDIR"]
				@configfile = conf["CONFIGFILE"]
				@mtime = Time::at(0)
				if !File::exist?(@configfile) then
					cgi = CGI.new("empty")
					sys = VikiWiki::Sys.new(@configfile, conf, cgi)
					sys.config({})
				end
				self.load
			end

			def filename; @configfile; end

			def load
				mtime = File::mtime(@configfile)
				if (@mtime < mtime) then
					script = File::read(@configfile)
					sysdef = @usrdef = "{}"
					if /@conf = (\{\n.*\n\})\n#{END_SDEF}/mn === script then
						sysdef = $1.untaint
					end
					sysdef.gsub!(/#{BGN_UDEF}(.*)#{END_UDEF}/mn) do
  						@usrdef = "{\n#{$1}}\n"; ""
					end
					@conf = init_conf(eval(sysdef))
					@conf["BASEDIR"] = @basedir
					@conf["CONFIGFILE"] = @configfile
					@conf.update(@scriptnames)
					@mtime = mtime
				end
				@conf.update(eval(@usrdef))
				return @conf
			end

			private

			def init_conf(conf)
				retconf = DEFAULT_CONFIG.dup; conf = conf.dup

				conf["BASEDIR"] ||= BASEDIR
				conf["STATICDIR"] ||= "#{conf['BASEDIR']}/static"
				conf["CONFIGFILE"] ||= "#{conf['BASEDIR']}/vikiwiki.conf"
				["BASEDIR", "STATICDIR", "CONFIGFILE"].each do |k|
					conf[k] = expand_path(conf[k])
				end

				conf["BASEURI"] ||= BASEURI
				conf["STATICURI"] ||= "#{conf["BASEURI"]}/static"
				["BASEURI", "STATICURI"].each do |k|
					conf[k] = normalize_uri(conf[k])
				end

				["WIKISTYLES", "LANGS"].each do |key|
					if Array === conf[key] then
						conf[key].each do |elem|
							retconf[key].include?(elem) || retconf[key].unshift(elem)
						end
						conf.delete(key)
					end
				end
				conf["WIKIFARM"] = false

				retconf.update(conf)
				return retconf
			end
		end

		class Repository
			include Util

			def initialize(destdir = nil)
				@destdir = expand_path(destdir || BASEDIR)
			end

			def self.install(destdir = nil)
				new(destdir).install
			end

			def install
				return if File.directory?(@destdir)
				FileUtils.mkdir_p(@destdir)
				FileUtils.chmod(0700, @destdir)
				Dir.chdir($VIKIWIKIMASTER) do
					Find::find(".") do |f|
						f.gsub!(%r|\./|, "")
						if FileTest.directory?(f) && !(/^\.\.?$/ =~ f) then
							FileUtils.mkdir_p("#{@destdir}/#{f}")
							FileUtils.chmod(0700, "#{@destdir}/#{f}")
						elsif FileTest.file?(f) then
							FileUtils.install(f, "#{@destdir}/#{File.dirname(f)}", {:mode => 0600})
						end
					end
				end
			end
		end

		############
		# class UserDB.
		# a passwd object for ::WEBrick::HTTPAuth::BasicAuth.

		class UserDB
			include ::WEBrick::HTTPAuth::UserDB
			KEYLEN = 16

			def initialize(etcdir)
				@path = "#{etcdir}/passwd"
				@auth_type = ::WEBrick::HTTPAuth::BasicAuth
				@passwd = Hash::new([]); @mtime = Time::at(0)
			end

			def reload
				if File::file?(@path) && \
					 ((mtime = File::mtime(@path)) > @mtime) then
					File::foreach(@path) do |line|
						user, pass, salt = line.split(":")
						@passwd[user] = [pass, salt]
					end
					@mtime = mtime
				end
			end

			def get_passwd(realm, user, reload_db)
				reload_db && reload(); return @passwd[user][0]
			end

			def set_passwd(realm, user, pass)
				salt = ::WEBrick::Utils::random_string(KEYLEN)
				pass = pass.crypt(salt)
				@passwd[user] = [pass, salt]
				text = @passwd.inject("") do |text, (user, (pass, salt))|
					text << "#{user}:#{pass}:#{salt}\n"; text
				end
				File::fwrite(@path, text); return pass
			end
		end

		############
		# WEBrick servlet module.
		############

		module ServletM
			include Util

			def self.included(klass)
				scriptname = klass::SCRIPTNAME
				klass.module_eval(<<-EOEVAL)

				Instances = Hash.new
				InstancesMutex = Mutex.new

				def self.mounted(dir, args = {})
					args = args.to_args
					args["#{scriptname}"] = dir
					return dir, args
				end

				def self.get_instance(server, args)
					InstancesMutex.synchronize do
						dir = args["#{scriptname}"]
						Instances[dir] ||= self.new(server, args)
					end
				end
				EOEVAL
			end

			def initialize(server, args = {})
				super
				args = args.to_args
				conf = check_repository(args)
				config_environment(conf)
				check_mount(server, conf)
				set_authenticate(conf)
			end

			def do_POST(request, response)
				(Thread.start do; ($SAFE < 1) && ($SAFE += 1)
					conf = @config.load
					filename = @config.filename
					check_authenticate(conf, request, response)
					cgi = request.to_cgi
					$VIKIWIKIOUT = response.to_out

					post_method_procedure(filename, conf, cgi)
				end).join
			end
			alias do_GET do_POST

			private

			############
			# check VikiWiki repository and configuration file.

			def check_repository(args)
				# check VikiWiki repository.
				Repository.install(args["BASEDIR"])
				# check VikiWiki config.
				@config = Config.new(args)
				return @config.load
			end

			############
			# set VikiWiki configuration hash and reset environment variable.

			def config_environment(conf)
				# add conf["BASEDIR"]/{,lib} to Ruby include path.
				$:.include?(conf["BASEDIR"]) || $:.unshift(conf["BASEDIR"])
				$:.include?("#{conf['BASEDIR']}/lib") || $:.unshift("#{conf['BASEDIR']}/lib")

				# set environment variable TMPDIR, TMP, PATH.
				$VIKIWIKITMP = ENV['TMPDIR'] = ENV['TMP'] = "#{conf['BASEDIR']}/tmp"
				ENV["PATH"] = "/usr/bin:/usr/sbin:/bin:/sbin"
			end

			############
			# check if conf["BASEDIR"]/{theme,img,script}
			# and conf["STATICDIR"] is mounted.

			def check_mount(server, conf)
				themeuri = (conf["THEMEURI"] || "#{conf["BASEURI"]}/theme")
				themedir = (conf["THEMEDIR"] || "#{conf["BASEDIR"]}/theme")
				check_mount_uri(server, themeuri, themedir)

				imguri = (conf["IMGURI"] || "#{conf["BASEURI"]}/img")
				imgdir = (conf["IMGDIR"] || "#{conf["BASEDIR"]}/img")
				check_mount_uri(server, imguri, imgdir)

				scripturi = "#{conf["BASEURI"]}/script"
				scriptdir = "#{conf["BASEDIR"]}/script"
				check_mount_uri(server, scripturi, scriptdir)

				check_mount_uri(server, conf["STATICURI"], conf["STATICDIR"])
			end

			def check_mount_uri(server, uri, dir)
				uri = normalize_uri(uri); dir = expand_path(dir)
				if !already_mounted_uri?(server, uri, dir) then
					server.mount(uri, ::WEBrick::HTTPServlet::FileHandler, dir)
				end
			end

			def already_mounted_uri?(server, uri, dir)
				servlet, options, _, _ = server.search_servlet(uri)
				(servlet == ::WEBrick::HTTPServlet::FileHandler) && (dir == options[0])
			end

			############
			# set authentication.

			def set_authenticate(conf)
				realm = conf["AUTHREALM"]
				etcdir = conf["ETCDIR"] || "#{conf["BASEDIR"]}/etc"
				userdb = UserDB::new(etcdir)

				if !userdb.get_passwd(realm, conf["ADMIN"], true) then
					userdb.set_passwd(realm, conf["ADMIN"], conf["ADMIN"])
				end

				@auth = ::WEBrick::HTTPAuth::BasicAuth.new(
					:Realm => realm, :UserDB => userdb
				)
			end

			def check_authenticate(conf, request, response)
				return if (conf["AUTHTYPE"] != "www")
				if conf["AUTHREALM"] != @auth.realm then
					realm = conf["AUTHREALM"]; userdb = @auth.userdb
					@auth = ::WEBrick::HTTPAuth::BasicAuth.new(:Realm => realm, :UserDB => userdb)
				end
				@auth.authenticate(request, response)
			end

			############
			# This method is called by do_POST.

			def post_method_procedure(filename, conf, cgi); end
		end

		class Servlet < ::WEBrick::HTTPServlet::AbstractServlet
			SCRIPTNAME = "SCRIPT_NAME"
			include ServletM

			private

			############
			# This method is called by do_POST.

			def post_method_procedure(filename, conf, cgi)
				VikiWiki::Web.new(VikiWiki::Sys.new(filename, conf, cgi))
			end
		end

		module PluginHook
			class Servlet < ::WEBrick::HTTPServlet::AbstractServlet
				SCRIPTNAME = "PLUGIN_HOOK_SCRIPT"
				include VikiWiki::WEBrick::ServletM

				undef do_GET

				private

				def check_mount(server, conf); end

				############
				# This method is called by do_POST.

				def post_method_procedure(filename, conf, cgi)
					VikiWiki::PluginHook.new(filename, conf, cgi, false)
				end
			end
		end

		module TrackBack
			class Servlet < ::WEBrick::HTTPServlet::AbstractServlet
				SCRIPTNAME = "TB_CGI"
				include VikiWiki::WEBrick::ServletM

				undef do_GET

				private

				def check_mount(server, conf); end

				############
				# This method is called by do_POST.

				def post_method_procedure(filename, conf, cgi)
					VikiWiki::TrackBack.new(filename, conf, cgi, false)
				end
			end
		end

		module XMLRPC
			class Servlet < ::XMLRPC::WEBrickServlet
				SCRIPTNAME = "XMLRPC_CGI"
				include VikiWiki::WEBrick::ServletM

				def initialize(server, args)
					super
					add_xmlrpc_handler
				end

				def service(request, response)
					(Thread.start do; ($SAFE < 1) && ($SAFE += 1)
						conf = @config.load
						filename = @config.filename
						check_authenticate(conf, request, response)
						cgi = request.to_cgi
						@xmlrpc.sys_setup(filename, conf, cgi)
						super
					end).join
				end
				undef do_GET; undef do_POST

				private

				def check_mount(server, conf); end

				def add_xmlrpc_handler
					@xmlrpc = VikiWiki::XMLRPC.new
					self.add_handler(VikiWiki::XMLRPC::INTERFACE, @xmlrpc)
					self.add_introspection
				end
			end
		end

		############
		# VikiWiki-WEBrick server class.

		class Server
			include Util

			def initialize(args = {})
				@args = args.to_args; @server = nil
			end

			############
			# Run VikiWiki server daemon.
			def self.run(args = {}); new(args).run; end
			def self.install(args = {}); new(args).install; end

			def run; fork {fork {start_daemon}}; end
			def install
				Repository.install(@args["BASEDIR"])
				Config.new(@args)
			end

			private

			def start_daemon
				basedir = (@args["BASEDIR"] ||= BASEDIR)
				Repository.install(basedir)
				open("#{basedir}/vikiwiki.log", "a") do |logfd|
					STDOUT.reopen(logfd); STDERR.reopen(logfd)
					start_server
				end
			end

			def start_server
				port = (@args["PORT"] || 10080); @args.delete("PORT")
				scriptname = (@args["SCRIPT_NAME"] ||= SCRIPT_NAME)
				xmlrpc_cgi = (@args["XMLRPC_CGI"] ||= XMLRPC_CGI)

				@server = ::WEBrick::HTTPServer.new(:Port => port)
				@server.mount(scriptname, VikiWiki::WEBrick::Servlet, @args)
				@server.mount(xmlrpc_cgi, VikiWiki::WEBrick::XMLRPC::Servlet, @args)

				trap("INT"){ @server.shutdown }; trap("TERM"){ @server.shutdown }
				@server.start
			end
		end
	end

	############
	# VikiWiki::Servlet == VikiWiki::WEBrick::Servlet
	Servlet = WEBrick::Servlet

	############
	# VikiWiki::WEBrickServer == VikiWiki::WEBrick::Server
	WEBrickServer = WEBrick::Server
end

############
# VikiWikiWEBrick == VikiWiki::WEBrick::Server

VikiWikiWEBrick = VikiWiki::WEBrick::Server

# ----------------------------------------------------------

if $0 == "-e" then
	def run(args = {})
		print "Start VikiWiki server...\n"
		VikiWiki::WEBrick::Server::run(args)
	end

	def install(args = {})
		print "Install VikiWiki repository...\n"
		VikiWiki::WEBrick::Server::install(args)
	end
end

# ----------------------------------------------------------

($0 == __FILE__) && VikiWiki::WEBrick::Server::run()

# ==========================================================
